import torch
import pytest
from idinn.sourcing_model import SingleSourcingModel, DualSourcingModel


@pytest.fixture
def single_sourcing_model():
    """
    Single sourcing model at default state.
    """
    return SingleSourcingModel(
        lead_time=2,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
    )


def test_single_sourcing_model_get_lead_time(
    single_sourcing_model: SingleSourcingModel,
):
    assert single_sourcing_model.get_lead_time() == 2


def test_single_sourcing_model_get_init_inventory(
    single_sourcing_model: SingleSourcingModel,
):
    assert torch.all(
        torch.eq(single_sourcing_model.get_init_inventory(), torch.tensor([10.0]))
    )


def test_single_sourcing_model_get_past_orders(
    single_sourcing_model: SingleSourcingModel,
):
    assert torch.all(
        torch.eq(single_sourcing_model.get_past_orders(), torch.zeros(3, 3))
    )


def test_single_sourcing_model_get_past_inventories(
    single_sourcing_model: SingleSourcingModel,
):
    assert torch.all(
        torch.eq(
            single_sourcing_model.get_past_inventories(),
            torch.tensor([10.0]).repeat(3, 3),
        )
    )


def test_single_sourcing_model_get_current_inventory(
    single_sourcing_model: SingleSourcingModel,
):
    assert torch.all(
        torch.eq(
            single_sourcing_model.get_current_inventory(),
            torch.tensor([[10.0], [10.0], [10.0]]),
        )
    )


def test_single_sourcing_model_get_cost(single_sourcing_model: SingleSourcingModel):
    assert torch.all(
        torch.eq(single_sourcing_model.get_cost(), torch.tensor([[5.0], [5.0], [5.0]]))
    )


@pytest.fixture
def single_sourcing_model_fixed_demand():
    """
    Single sourcing model with fixed demand at 1.
    """
    return SingleSourcingModel(
        lead_time=1,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
        demand=[1, 1, 1, 1, 1],
    )


def test_single_sourcing_model_order(
    single_sourcing_model_fixed_demand: SingleSourcingModel,
):
    single_sourcing_model_fixed_demand.order(torch.tensor([[1.0], [2.0], [3.0]]))
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_past_orders(),
            torch.cat([torch.zeros(3, 2), torch.tensor([[1.0], [2.0], [3.0]])], dim=1),
        )
    )
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_past_inventories(),
            torch.cat(
                [
                    torch.tensor([10.0]).repeat(3, 2),
                    torch.tensor([[9.0], [9.0], [9.0]]),
                ],
                dim=1,
            ),
        )
    )
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_current_inventory(),
            torch.tensor([[9.0], [9.0], [9.0]]),
        )
    )
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_cost(),
            torch.tensor([[4.5], [4.5], [4.5]]),
        )
    )

    single_sourcing_model_fixed_demand.order(torch.tensor([[1.0], [2.0], [3.0]]))
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_past_orders(),
            torch.cat(
                [torch.zeros(3, 2), torch.tensor([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])],
                dim=1,
            ),
        )
    )
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_current_inventory(),
            torch.tensor([[9.0], [10.0], [11.0]]),
        )
    )
    assert torch.all(
        torch.eq(
            single_sourcing_model_fixed_demand.get_cost(),
            torch.tensor([[4.5], [5.0], [5.5]]),
        )
    )


@pytest.fixture
def dual_sourcing_model():
    """
    Dual sourcing model at default state.
    """
    return DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=1,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
    )


def test_dual_sourcing_model_get_expedited_lead_time(
    dual_sourcing_model: DualSourcingModel,
):
    assert dual_sourcing_model.get_expedited_lead_time() == 1


def test_dual_sourcing_model_get_regular_lead_time(
    dual_sourcing_model: DualSourcingModel,
):
    assert dual_sourcing_model.get_regular_lead_time() == 2


def test_dual_sourcing_model_get_init_inventory(dual_sourcing_model: DualSourcingModel):
    assert torch.all(
        torch.eq(dual_sourcing_model.get_init_inventory(), torch.tensor([10.0]))
    )


def test_dual_sourcing_model_get_past_regular_orders(
    dual_sourcing_model: DualSourcingModel,
):
    assert torch.all(
        torch.eq(dual_sourcing_model.get_past_regular_orders(), torch.zeros(3, 3))
    )


def test_dual_sourcing_model_get_past_expedited_orders(
    dual_sourcing_model: DualSourcingModel,
):
    assert torch.all(
        torch.eq(dual_sourcing_model.get_past_expedited_orders(), torch.zeros(3, 3))
    )


def test_dual_sourcing_model_get_past_inventories(
    dual_sourcing_model: DualSourcingModel,
):
    assert torch.all(
        torch.eq(
            dual_sourcing_model.get_past_inventories(),
            torch.tensor([10.0]).repeat(3, 3),
        )
    )


def test_dual_sourcing_model_get_current_inventory(
    dual_sourcing_model: DualSourcingModel,
):
    assert torch.all(
        torch.eq(
            dual_sourcing_model.get_current_inventory(),
            torch.tensor([[10.0], [10.0], [10.0]]),
        )
    )


def test_dual_sourcing_model_get_cost(dual_sourcing_model: DualSourcingModel):
    assert torch.all(
        torch.eq(
            dual_sourcing_model.get_cost(0, 0), torch.tensor([[5.0], [5.0], [5.0]])
        )
    )


@pytest.fixture
def dual_sourcing_model_fixed_demand():
    """
    Dual sourcing model with fixed demand at 1.
    """
    return DualSourcingModel(
        regular_lead_time=1,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
        demand=[1, 1, 1, 1, 1],
    )


def test_dual_sourcing_model_order(
    dual_sourcing_model_fixed_demand: DualSourcingModel,
):
    dual_sourcing_model_fixed_demand.order(
        regular_q=torch.tensor([[1.0], [2.0], [3.0]]),
        expedited_q=torch.tensor([[4.0], [5.0], [6.0]]),
    )
    assert torch.all(
        torch.eq(
            dual_sourcing_model_fixed_demand.get_past_regular_orders(),
            torch.cat([torch.zeros(3, 2), torch.tensor([[1.0], [2.0], [3.0]])], dim=1),
        )
    )
    assert torch.all(
        torch.eq(
            dual_sourcing_model_fixed_demand.get_past_expedited_orders(),
            torch.cat([torch.zeros(3, 2), torch.tensor([[4.0], [5.0], [6.0]])], dim=1),
        )
    )
    assert torch.all(
        torch.eq(
            dual_sourcing_model_fixed_demand.get_past_inventories(),
            torch.cat(
                [
                    torch.tensor([10.0]).repeat(3, 2),
                    torch.tensor([[13.0], [14.0], [15.0]]),
                ],
                dim=1,
            ),
        )
    )
    assert torch.all(
        torch.eq(
            dual_sourcing_model_fixed_demand.get_current_inventory(),
            torch.tensor([[13.0], [14.0], [15.0]]),
        )
    )
    assert torch.all(
        torch.eq(
            dual_sourcing_model_fixed_demand.get_cost(regular_q=0, expedited_q=1),
            torch.tensor([[26.5], [27.0], [27.5]]),
        )
    )



import torch


class BaseSourcingModel:
    def __init__(
        self,
        holding_cost,
        shortage_cost,
        init_inventory,
        batch_size,
        lead_time=None,
        regular_lead_time=None,
        expedited_lead_time=None,
        regular_order_cost=None,
        expedited_order_cost=None,
        demand=None,
        demand_distribuion="uniform",
        demand_low=1,
        demand_high=4,
    ):
        self.holding_cost = holding_cost
        self.shortage_cost = shortage_cost
        self.init_inventory = torch.tensor(
            [init_inventory], requires_grad=True, dtype=torch.float
        )
        self.batch_size = batch_size
        self.lead_time = lead_time
        self.regular_lead_time = regular_lead_time
        self.expedited_lead_time = expedited_lead_time
        self.regular_order_cost = regular_order_cost
        self.expedited_order_cost = expedited_order_cost

        if demand is not None:
            self.demand_iterator = iter(demand)
        elif demand_distribuion is not None:
            if demand_distribuion == "uniform":
                self.demand_iterator = torch.distributions.Uniform(
                    low=demand_low, high=demand_high + 1
                )
            else:
                raise ValueError("Only `uniform` distribution is supported.")
        else:
            raise ValueError("`demand` or `demand_distribution` must be provided.")
        self.reset()

    def reset(
        self,
        batch_size=None,
    ):
        if batch_size is not None and self.batch_size != batch_size:
            self.batch_size = batch_size

        if self.lead_time is not None:
            self.past_orders = torch.zeros(self.batch_size, self.lead_time + 1)
            self.past_inventories = self.get_init_inventory().repeat(
                self.batch_size, self.lead_time + 1
            )
        elif (
            self.regular_lead_time is not None and self.expedited_lead_time is not None
        ):
            if self.regular_lead_time < self.expedited_lead_time:
                raise ValueError(
                    "`regular_lead_time` must be greater than or equal to expedited_lead_time."
                )
            self.past_regular_orders = torch.zeros(
                self.batch_size, self.regular_lead_time + 1
            )
            self.past_expedited_orders = torch.zeros(
                self.batch_size, self.regular_lead_time + 1
            )
            self.past_inventories = self.get_init_inventory().repeat(
                self.batch_size, self.regular_lead_time + 1
            )
        else:
            raise ValueError(
                "Either `lead_time` or (`regular_lead_time` and `expedited_lead_time`) must be provided."
            )

    def get_init_inventory(self):
        init_inventory = (
            self.init_inventory - torch.frac(self.init_inventory).clone().detach()
        )
        return init_inventory

    def get_past_inventories(self):
        return self.past_inventories

    def get_current_inventory(self):
        # Keep dim when slicing past_inventories
        # https://stackoverflow.com/questions/57237352/why-does-torch-slice-lose-dimension
        # https://discuss.pytorch.org/t/solved-simple-question-about-keep-dim-when-slicing-the-tensor/9280
        return self.past_inventories[:, [-1]]

    def generate_demand(self):
        """
        Generate demands depending on the type of demand_iterator.
        """
        if isinstance(self.demand_iterator, torch.distributions.Uniform):
            current_demand = self.demand_iterator.sample((self.batch_size, 1)).int()
        else:
            try:
                current_demand = next(self.demand_iterator)
            except StopIteration:
                raise ValueError(
                    "`demand` is not long enough for the overall sourcing periods."
                )
            # Ensure demand is a tensor of non-negative integers and the shape conforms to (batch_size,1)
            if not isinstance(current_demand, torch.Tensor):
                current_demand = (
                    torch.tensor(current_demand).int().repeat(self.batch_size, 1)
                )
                current_demand = torch.clamp(current_demand, min=0)
        return current_demand


class SingleSourcingModel(BaseSourcingModel):
    def __init__(
        self,
        lead_time,
        holding_cost,
        shortage_cost,
        init_inventory,
        batch_size=1,
        demand=None,
        demand_distribuion="uniform",
        demand_low=1,
        demand_high=4,
    ):
        """
        Parameters
        ----------
        lead_time : float
            The lead time for orders.
        holding_cost : float
            The cost of holding inventory.
        shortage_cost : float
            The cost of inventory shortage.
        init_inventory : float
            The initial inventory.
        batch_size : int, optional
            The batch size for orders. default is 1.
        demand : Iterable, torch.Tensor, np.array, or pd.Series, optional
            The array for outputting values of demands.
        demand_distribuion : str, default is `uniform`.
            Distribution for generated demands when `demand` is not specified.
        demand_low : int, default is 1.
            Lower inclusive bound for generated demands when `demand` is not specified.
        demand_high : int, default is 4.
            Higher inclusive bound for generated demands when `demand` is not specified.
        """
        super().__init__(
            lead_time=lead_time,
            holding_cost=holding_cost,
            shortage_cost=shortage_cost,
            init_inventory=init_inventory,
            batch_size=batch_size,
            demand=demand,
            demand_distribuion=demand_distribuion,
            demand_low=demand_low,
            demand_high=demand_high,
        )

    def get_lead_time(self):
        return self.lead_time

    def get_past_orders(self):
        return self.past_orders

    def get_cost(self):
        cost = self.holding_cost * torch.relu(
            self.get_current_inventory()
        ) + self.shortage_cost * torch.relu(-self.get_current_inventory())
        return cost

    def order(self, q, seed=None):
        """
        Orders items to the inventory and update the inventory with generated demands.

        Parameters
        ----------
        q : int, or torch.Tensor
            The quantity of items to order.
        seed : int, optional
            Random seed for reproducibility.
        """
        if seed is not None:
            torch.manual_seed(seed)
        # Current orders are added to past_orders
        self.past_orders = torch.cat([self.past_orders, q], dim=1)
        # Past orders arrived
        arrived_order = self.past_orders[:, [-1 - self.lead_time]]
        # Generate current demand
        current_demand = self.generate_demand()
        # Update inventory
        current_inventory = (
            self.get_current_inventory() + arrived_order - current_demand
        )
        # Current inventories are added to past_inventories
        self.past_inventories = torch.cat(
            [self.past_inventories, current_inventory], dim=1
        )


class DualSourcingModel(BaseSourcingModel):
    def __init__(
        self,
        regular_lead_time,
        expedited_lead_time,
        regular_order_cost,
        expedited_order_cost,
        holding_cost,
        shortage_cost,
        init_inventory,
        batch_size=1,
        demand=None,
        demand_distribuion="uniform",
        demand_low=1,
        demand_high=4,
    ):
        """
        Parameters
        ----------
        regular_lead_time : float
            The lead time for regular orders.
        expedited_lead_time : float
            The lead time for expedited orders.
        regular_order_cost : float
            The cost of placing a regular order.
        expedited_order_cost : float
            The cost of placing an expedited order.
        holding_cost : float
            The cost of holding inventory.
        shortage_cost : float
            The cost of shortage.
        init_inventory : float
            The initial inventory.
        batch_size : int, default is 1
            The batch size for orders.
        demand : Iterable, torch.Tensor, np.array, or pd.Series, optional
            The array for outputting values of demands.
        demand_distribuion : str, default is `uniform`.
            Distribution for generated demand when `demand` is not specified.
        demand_low : int, default is 1.
            Lower bound for generated demand when `demand` is not specified.
        demand_high : int, default is 4.
            Higher bound for generated demand when `demand` is not specified.
        """
        super().__init__(
            regular_lead_time=regular_lead_time,
            expedited_lead_time=expedited_lead_time,
            regular_order_cost=regular_order_cost,
            expedited_order_cost=expedited_order_cost,
            holding_cost=holding_cost,
            shortage_cost=shortage_cost,
            init_inventory=init_inventory,
            batch_size=batch_size,
            demand=demand,
            demand_distribuion=demand_distribuion,
            demand_low=demand_low,
            demand_high=demand_high,
        )

    def get_past_regular_orders(self):
        return self.past_regular_orders

    def get_past_expedited_orders(self):
        return self.past_expedited_orders

    def get_regular_lead_time(self):
        return self.regular_lead_time

    def get_expedited_lead_time(self):
        return self.expedited_lead_time

    def get_cost(self, regular_q, expedited_q):
        cost = (
            self.regular_order_cost * regular_q
            + self.expedited_order_cost * expedited_q
            + self.holding_cost * torch.relu(self.get_current_inventory())
            + self.shortage_cost * torch.relu(-self.get_current_inventory())
        )
        return cost

    def order(self, regular_q, expedited_q, seed=None):
        """
        Orders items to the inventory and update the inventory with generated demands.

        Parameters
        ----------
        regular_q : int, or torch.Tensor
            The quantity of items to order from the regular supplier.
        expedited_q : int, or torch.Tensor
            The quantity of items to order from the expedited supplier.    
        seed : int, optional
            Random seed for reproducibility.
        """
        if seed is not None:
            torch.manual_seed(seed)
        if not isinstance(regular_q, torch.Tensor):
            regular_q = torch.tensor([[regular_q]])
        if not isinstance(expedited_q, torch.Tensor):
            expedited_q = torch.tensor([[expedited_q]])
        # Current regular order are added to past_regular_orders
        self.past_regular_orders = torch.cat(
            [self.past_regular_orders, regular_q], dim=1
        )
        # Current expedited order are added to past_expedited_orders
        self.past_expedited_orders = torch.cat(
            [self.past_expedited_orders, expedited_q], dim=1
        )
        # Past regular orders arrived
        arrived_regular_orders = self.past_regular_orders[
            :, [-1 - self.regular_lead_time]
        ]
        # Past expedited orders arrived
        arrived_expedited_orders = self.past_expedited_orders[
            :, [-1 - self.expedited_lead_time]
        ]
        # Generate current demand
        current_demand = self.generate_demand()
        # Update inventory
        current_inventory = (
            self.get_current_inventory()
            + arrived_expedited_orders
            + arrived_regular_orders
            - current_demand
        )
        self.past_inventories = torch.cat(
            [self.past_inventories, current_inventory], dim=1
        )



import matplotlib.pyplot as plt
import numpy as np
import torch


class NeuralControllerMixIn():
    def save(self, checkpoint_path):
        torch.save(self.state_dict(), checkpoint_path)

    def load(self, checkpoint_path):
        self.load_state_dict(torch.load(checkpoint_path))


class SingleSourcingNeuralController(torch.nn.Module, NeuralControllerMixIn):
    """
    SingleSourcingNeuralController is a neural network-based controller for inventory optimization in a single-sourcing scenario.

    Parameters
    ----------
    hidden_layers : list, default is [2]
        List of integers representing the number of units in each hidden layer.
    activation : torch.nn.Module, default is torch.nn.CELU(alpha=1)
        Activation function to be used in the hidden layers.

    Attributes
    ----------
    hidden_layers : list
        List of integers representing the number of units in each hidden layer.
    activation : torch.nn.Module
        Activation function used in the hidden layers.
    stack : torch.nn.Sequential
        Sequential stack of linear layers and activation functions.

    Methods
    -------
    init_layers(lead_time)
        Initialize the layers of the neural network based on the lead time.
    forward(current_inventory, past_orders)
        Perform forward pass through the neural network.
    get_total_cost(sourcing_model, sourcing_periods, seed=None)
        Calculate the total cost over a given number of sourcing periods.
    train(sourcing_model, sourcing_periods, epochs, ...)
        Train the neural network controller using the sourcing model and specified parameters.
    simulate(sourcing_model, sourcing_periods)
        Simulate the inventory and order quantities over a given number of sourcing periods.
    plot(sourcing_model, sourcing_periods)
        Plot the inventory and order quantities over a given number of sourcing periods.
    """

    def __init__(self, hidden_layers=[2], activation=torch.nn.CELU(alpha=1)):
        super().__init__()
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.lead_time = None
        self.stack = None

    def init_layers(self, lead_time):
        """
        Initialize the layers of the neural network based on the lead time.

        Parameters
        ----------
        lead_time : int
            The lead time for sourcing.

        Returns
        -------
        None
        """
        self.lead_time = lead_time
        architecture = [
            torch.nn.Linear(lead_time + 1, self.hidden_layers[0]),
            self.activation,
        ]
        for i in range(len(self.hidden_layers)):
            if i < len(self.hidden_layers) - 1:
                architecture += [
                    torch.nn.Linear(self.hidden_layers[i], self.hidden_layers[i + 1]),
                    self.activation,
                ]
        architecture += [
            torch.nn.Linear(self.hidden_layers[-1], 1, bias=False),
            torch.nn.ReLU(),
        ]
        self.stack = torch.nn.Sequential(*architecture)

    def forward(
        self,
        current_inventory,
        past_orders,
    ):
        """
        Perform forward pass through the neural network.

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory levels.
        past_orders : int, or torch.Tensor
            Past order quantities.

        Returns
        -------
        torch.Tensor
            Order quanty calculated by the neural network.
        """
        if not isinstance(current_inventory, torch.Tensor):
            current_inventory = torch.tensor([[current_inventory]], dtype=torch.float32)
        if not isinstance(past_orders, torch.Tensor):
            past_orders = torch.tensor([past_orders], dtype=torch.float32)
        if self.lead_time > 0:
            h = torch.cat([current_inventory, past_orders[:, -self.lead_time :]], dim=1)
        h = self.stack(h)
        q = h - torch.frac(h).clone().detach()
        return q

    def get_total_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the total cost over a given number of sourcing periods.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model to be used for cost calculation.
        sourcing_periods : int
            The number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        numpy.ndarray
            Total cost over the sourcing periods.
        """
        if seed is not None:
            torch.manual_seed(seed)

        if self.lead_time is None:
            self.init_layers(sourcing_model.get_lead_time())

        total_cost = 0
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_orders = sourcing_model.get_past_orders()
            q = self.forward(current_inventory, past_orders)
            sourcing_model.order(q)
            current_cost = sourcing_model.get_cost()
            total_cost += current_cost.mean()
        return total_cost

    def train(
        self,
        sourcing_model,
        sourcing_periods,
        epochs,
        validation_sourcing_periods=None,
        lr_init_inventory=1e-1,
        lr_parameters=3e-3,
        seed=None,
        tensorboard_writer=None,
    ):
        """
        Train the neural network controller using the sourcing model and specified parameters.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model to be used for training.
        sourcing_periods : int
            The number of sourcing periods for training.
        epochs : int
            The number of training epochs.
        validation_sourcing_periods : int, optional
            The number of sourcing periods for validation.
        lr_init_inventory : float, default is 1e-1
            Learning rate for initial inventory.
        lr_parameters : float, default is 3e-3
            Learning rate for updating neural network parameters. 
        seed : int, optional
            Random seed for reproducibility.
        tensorboard_writer : tensorboard.SummaryWriter, optional
            Tensorboard writer for logging.
        """
        if seed is not None:
            torch.manual_seed(seed)

        optimizer_init_inventory = torch.optim.RMSprop(
            [sourcing_model.init_inventory], lr=lr_init_inventory
        )
        optimizer_parameters = torch.optim.RMSprop(self.parameters(), lr=lr_parameters)
        min_cost = np.inf

        for epoch in range(epochs):
            # Clear grad cache
            optimizer_parameters.zero_grad()
            optimizer_init_inventory.zero_grad()
            # Reset the sourcing model with the learned init inventory
            sourcing_model.reset()
            total_cost = self.get_total_cost(sourcing_model, sourcing_periods)
            total_cost.backward()
            # Gradient descend
            if epoch % 3 == 0:
                optimizer_init_inventory.step()
            else:
                optimizer_parameters.step()
            # Save the best model
            if validation_sourcing_periods is not None and epoch % 10 == 0:
                eval_cost = self.get_total_cost(
                    sourcing_model, validation_sourcing_periods
                )
                if eval_cost < min_cost:
                    min_cost = eval_cost
                    best_state = self.state_dict()
            else:
                if total_cost < min_cost:
                    min_cost = total_cost
                    best_state = self.state_dict()
            # Log train loss
            if tensorboard_writer is not None:
                tensorboard_writer.add_scalar(
                    "Avg. cost per period/train", total_cost / sourcing_periods, epoch
                )
                # Log evaluation loss
                if validation_sourcing_periods is not None and epoch % 10 == 0:
                    eval_cost = self.get_total_cost(
                        sourcing_model, validation_sourcing_periods
                    )
                    tensorboard_writer.add_scalar(
                        "Avg. cost per period/eval",
                        eval_cost / validation_sourcing_periods,
                        epoch,
                    )
                tensorboard_writer.flush()
        # Load the best model
        self.load_state_dict(best_state)

    def simulate(self, sourcing_model, sourcing_periods, seed=None):
        """
        Simulate the inventory and order quantities over a given number of sourcing periods.

        Parameters
        ----------
        sourcing_model : SingleSourcingModel
            The sourcing model to be used for simulation.
        sourcing_periods : int
            The number of sourcing periods for simulation.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        tuple
            A tuple containing the past inventories and past orders as numpy arrays.
        """
        if seed is not None:
            torch.manual_seed(seed)
        sourcing_model.reset(batch_size=1)
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_orders = sourcing_model.get_past_orders()
            q = self.forward(current_inventory, past_orders)
            sourcing_model.order(q)
        past_inventories = sourcing_model.get_past_inventories()[0, :].detach().numpy()
        past_orders = sourcing_model.get_past_orders()[0, :].detach().numpy()
        return past_inventories, past_orders

    def plot(self, sourcing_model, sourcing_periods):
        """
        Plot the inventory and order quantities over a given number of sourcing periods.

        Parameters
        ----------
        sourcing_model : SingleSourcingModel
            The sourcing model to be used for plotting.
        sourcing_periods : int
            The number of sourcing periods for plotting.
        """
        past_inventories, past_orders = self.simulate(
            sourcing_model=sourcing_model, sourcing_periods=sourcing_periods
        )
        fig, ax = plt.subplots(ncols=2, figsize=(10, 4))

        ax[0].step(range(sourcing_periods), past_inventories[-sourcing_periods:])
        ax[0].yaxis.get_major_locator().set_params(integer=True)
        ax[0].set_title("Inventory")
        ax[0].set_xlabel("Period")
        ax[0].set_ylabel("Quantity")

        ax[1].step(range(sourcing_periods), past_orders[-sourcing_periods:])
        ax[1].yaxis.get_major_locator().set_params(integer=True)
        ax[1].set_title("Order")
        ax[1].set_xlabel("Period")
        ax[1].set_ylabel("Quantity")


class DualSourcingNeuralController(torch.nn.Module, NeuralControllerMixIn):
    """
    DualSourcingNeuralController is a neural network controller for dual sourcing inventory optimization.

    Parameters
    ----------
    hidden_layers : list, default is [128, 64, 32, 16, 8, 4]
        List of integers specifying the sizes of hidden layers.
    activation : torch.nn.Module, default is torch.nn.CELU(alpha=1)
        Activation function to be used in the hidden layers.
    compressed : bool, default is False
        Flag indicating whether the input is compressed.

    Attributes
    ----------
    hidden_layers : list
        List of integers specifying the sizes of hidden layers.
    activation : torch.nn.Module
        Activation function to be used in the hidden layers.
    compressed : bool
        Flag indicating whether the input is compressed.
    regular_lead_time : int
        Regular lead time.
    expedited_lead_time : int
        Expedited lead time.
    stack : torch.nn.Sequential
        Sequential stack of linear layers and activation functions.

    Methods
    -------
    init_layers(regular_lead_time, expedited_lead_time)
        Initialize the layers of the neural network.
    forward(current_inventory, past_orders)
        Forward pass of the neural network.
    get_total_cost(sourcing_model, sourcing_periods, seed=None)
        Calculate the total cost of the sourcing model.
    train(sourcing_model, sourcing_periods, epochs, ...)
        Trains the neural network controller using the sourcing model and specified parameters.
    simulate(sourcing_model, sourcing_periods, seed=None)
        Simulate the sourcing model using the neural network.
    plot(sourcing_model, sourcing_periods)
        Plot the inventory and order quantities.
    """

    def __init__(
        self,
        hidden_layers=[128, 64, 32, 16, 8, 4],
        activation=torch.nn.CELU(alpha=1),
        compressed=False,
    ):
        super().__init__()
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.compressed = compressed
        self.lead_time = None
        self.stack = None

    def init_layers(self, regular_lead_time, expedited_lead_time):
        """
        Initialize the layers of the neural network.

        Parameters
        ----------
        regular_lead_time : int
            Regular lead time.
        expedited_lead_time : int
            Expedited lead time.
        """
        self.regular_lead_time = regular_lead_time
        self.expedited_lead_time = expedited_lead_time
        if self.compressed:
            input_length = regular_lead_time + expedited_lead_time
        else:
            input_length = regular_lead_time + expedited_lead_time + 1

        architecture = [
            torch.nn.Linear(input_length, self.hidden_layers[0]),
            self.activation,
        ]
        for i in range(len(self.hidden_layers)):
            if i < len(self.hidden_layers) - 1:
                architecture += [
                    torch.nn.Linear(self.hidden_layers[i], self.hidden_layers[i + 1]),
                    self.activation,
                ]
        architecture += [
            torch.nn.Linear(self.hidden_layers[-1], 2),
            torch.nn.ReLU(),
        ]
        self.stack = torch.nn.Sequential(*architecture)

    def forward(self, current_inventory, past_regular_orders, past_expedited_orders):
        """
        Forward pass of the neural network.

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory.
        past_regular_orders : int, or torch.Tensor
            Past regular orders.
        past_expedited_orders : int, or torch.Tensor
            Past expedited orders.

        Returns
        -------
        regular_q : torch.Tensor
            Regular order quantity.
        expedited_q : torch.Tensor
            Expedited order quantity.
        """
        if not isinstance(current_inventory, torch.Tensor):
            current_inventory = torch.tensor([[current_inventory]], dtype=torch.float32)
        if not isinstance(past_regular_orders, torch.Tensor):
            past_regular_orders = torch.tensor([past_regular_orders], dtype=torch.float32)
        if not isinstance(past_expedited_orders, torch.Tensor):
            past_expedited_orders = torch.tensor([past_expedited_orders], dtype=torch.float32)

        if self.regular_lead_time > 0:
            if self.compressed:
                inputs = past_regular_orders[:, -self.regular_lead_time :]
                inputs[:, 0] += current_inventory
            else:
                inputs = torch.cat(
                    [
                        current_inventory,
                        past_regular_orders[:, -self.regular_lead_time :],
                    ],
                    dim=1,
                )

        if self.expedited_lead_time > 0:
            inputs = torch.cat(
                [inputs, past_expedited_orders[:, -self.expedited_lead_time :]], dim=1
            )

        h = self.stack(inputs)
        q = h - torch.frac(h).clone().detach()
        regular_q = q[:, [0]]
        expedited_q = q[:, [1]]
        return regular_q, expedited_q

    def get_total_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the total cost of the sourcing model.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        total_cost : torch.Tensor
            Total cost.
        """
        if seed is not None:
            torch.manual_seed(seed)

        if self.lead_time is None:
            self.init_layers(
                regular_lead_time=sourcing_model.get_regular_lead_time(),
                expedited_lead_time=sourcing_model.get_expedited_lead_time(),
            )
        
        total_cost = 0
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_regular_orders = sourcing_model.get_past_regular_orders()
            past_expedited_orders = sourcing_model.get_past_expedited_orders()
            regular_q, expedited_q = self.forward(
                current_inventory, past_regular_orders, past_expedited_orders
            )
            sourcing_model.order(regular_q, expedited_q)
            current_cost = sourcing_model.get_cost(regular_q, expedited_q)
            total_cost += current_cost.mean()
        return total_cost

    def train(
        self,
        sourcing_model,
        sourcing_periods,
        epochs,
        validation_sourcing_periods=None,
        lr_init_inventory=1e-1,
        lr_parameters=3e-3,
        seed=None,
        tensorboard_writer=None,
    ):
        """
        Train the neural network.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        epochs : int
            Number of training epochs.
        validation_sourcing_periods : int, optional
            Number of sourcing periods for validation.
        lr_init_inventory : float, default is 1e-1
            Learning rate for initial inventory.
        lr_parameters : float, default is 3e-3
            Learning rate for updating neural network parameters.
        seed : int, optional
            Random seed for reproducibility.
        tensorboard_writer : TensorBoard writer, optional
            TensorBoard writer for logging.
        """
        if seed is not None:
            torch.manual_seed(seed)

        optimizer_init_inventory = torch.optim.RMSprop(
            [sourcing_model.init_inventory], lr=lr_init_inventory
        )
        optimizer_parameters = torch.optim.RMSprop(self.parameters(), lr=lr_parameters)
        min_cost = np.inf

        for epoch in range(epochs):
            # Clear grad cache
            optimizer_init_inventory.zero_grad()
            optimizer_parameters.zero_grad()
            # Reset the sourcing model with the learned init inventory
            sourcing_model.reset()
            total_cost = self.get_total_cost(sourcing_model, sourcing_periods)
            total_cost.backward()
            # Perform gradient descend
            if epoch % 3 == 0:
                optimizer_init_inventory.step()
            else:
                optimizer_parameters.step()
            # Save the best model
            if validation_sourcing_periods is not None and epoch % 10 == 0:
                eval_cost = self.get_total_cost(
                    sourcing_model, validation_sourcing_periods
                )
                if eval_cost < min_cost:
                    min_cost = eval_cost
                    best_state = self.state_dict()
            else:
                if total_cost < min_cost:
                    min_cost = total_cost
                    best_state = self.state_dict()
            # Log train loss
            if tensorboard_writer is not None:
                tensorboard_writer.add_scalar(
                    "Avg. cost per period/train", total_cost / sourcing_periods, epoch
                )
                # Log evaluation loss
                tensorboard_writer.add_scalar(
                    "Avg. cost per period/eval",
                    eval_cost / validation_sourcing_periods,
                    epoch,
                )
                tensorboard_writer.flush()

        self.load_state_dict(best_state)

    def simulate(self, sourcing_model, sourcing_periods, seed=None):
        """
        Simulate the sourcing model using the neural network.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        past_inventories : list
            List of past inventories.
        past_regular_orders : list
            List of past regular orders.
        past_expedited_orders : list
            List of past expedited orders.

        """
        if seed is not None:
            torch.manual_seed(seed)
        sourcing_model.reset(batch_size=1)
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_regular_orders = sourcing_model.get_past_regular_orders()
            past_expedited_orders = sourcing_model.get_past_expedited_orders()
            regular_q, expedited_q = self.forward(
                current_inventory, past_regular_orders, past_expedited_orders
            )
            sourcing_model.order(regular_q, expedited_q)
        past_inventories = sourcing_model.get_past_inventories()[0, :].detach().numpy()
        past_regular_orders = (
            sourcing_model.get_past_regular_orders()[0, :].detach().numpy()
        )
        past_expedited_orders = (
            sourcing_model.get_past_expedited_orders()[0, :].detach().numpy()
        )
        return past_inventories, past_regular_orders, past_expedited_orders

    def plot(self, sourcing_model, sourcing_periods):
        """
        Plot the inventory and order quantities.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        """
        past_inventories, past_regular_orders, past_expedited_orders = self.simulate(
            sourcing_model=sourcing_model, sourcing_periods=sourcing_periods
        )
        fig, ax = plt.subplots(ncols=2, figsize=(10, 4))
        ax[0].step(range(sourcing_periods), past_inventories[-sourcing_periods:])
        ax[0].yaxis.get_major_locator().set_params(integer=True)
        ax[0].set_title("Inventory")
        ax[0].set_xlabel("Period")
        ax[0].set_ylabel("Quantity")

        ax[1].step(
            range(sourcing_periods),
            past_expedited_orders[-sourcing_periods:],
            label="Expedited Order",
        )
        ax[1].step(
            range(sourcing_periods),
            past_regular_orders[-sourcing_periods:],
            label="Regular Order",
        )
        ax[1].yaxis.get_major_locator().set_params(integer=True)
        ax[1].set_title("Order")
        ax[1].set_xlabel("Period")
        ax[1].set_ylabel("Quantity")
        ax[1].legend()


import pytest
import torch
from idinn.sourcing_model import SingleSourcingModel, DualSourcingModel
from idinn.controller import SingleSourcingNeuralController, DualSourcingNeuralController

@pytest.fixture
def single_sourcing_model():
    return SingleSourcingModel(
        lead_time=2,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=1,
        demand_distribuion="uniform",
        demand_low=1,
        demand_high=4
    )

@pytest.fixture
def single_sourcing_controller():
    single_sourcing_controller = SingleSourcingNeuralController(
        hidden_layers = [4, 2],
        activation = torch.nn.ReLU()
    )
    single_sourcing_controller.init_layers(lead_time=2)
    return single_sourcing_controller

def test_single_controller_init_layers(single_sourcing_controller: SingleSourcingNeuralController):
    assert len(single_sourcing_controller.stack) == 6

def test_single_controller_forward(single_sourcing_controller: SingleSourcingNeuralController):
    current_inventory = 10
    past_orders = [0, 0]
    q = single_sourcing_controller.forward(current_inventory, past_orders)
    assert q.shape == torch.Size([1, 1])

def test_single_controller_get_total_cost(single_sourcing_model: SingleSourcingModel, single_sourcing_controller: SingleSourcingNeuralController):
    total_cost = single_sourcing_controller.get_total_cost(single_sourcing_model, sourcing_periods=5, seed=42)
    assert total_cost.item() == 14.0

def test_single_controller_simulate(single_sourcing_model: SingleSourcingModel, single_sourcing_controller: SingleSourcingNeuralController):
    past_inventories, past_orders = single_sourcing_controller.simulate(single_sourcing_model, sourcing_periods=5, seed=42)
    assert past_inventories[0] == 10
    assert past_orders[0] == 0

def test_single_controller_train(single_sourcing_model: SingleSourcingModel, single_sourcing_controller: SingleSourcingNeuralController):
    single_sourcing_controller.train(single_sourcing_model, sourcing_periods=5, epochs=1)

def test_single_controller_plot(single_sourcing_model: SingleSourcingModel, single_sourcing_controller: SingleSourcingNeuralController):
    single_sourcing_controller.plot(single_sourcing_model, sourcing_periods=5)

@pytest.fixture
def dual_sourcing_model():
    """
    Dual sourcing model at default state.
    """
    return DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=1,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=1,
        demand_distribuion="uniform",
        demand_low=1,
        demand_high=4
    )

@pytest.fixture
def dual_sourcing_controller():
    dual_sourcing_controller = DualSourcingNeuralController(
        hidden_layers = [4, 2],
        activation = torch.nn.ReLU()
    )
    dual_sourcing_controller.init_layers(regular_lead_time=2, expedited_lead_time=1)
    return dual_sourcing_controller

def test_dual_controller_init_layers(dual_sourcing_controller: DualSourcingNeuralController):
    assert len(dual_sourcing_controller.stack) == 6
    assert dual_sourcing_controller.stack[-2].out_features == 2

def test_dual_controller_forward(dual_sourcing_controller: DualSourcingNeuralController):
    current_inventory = 10
    past_regular_orders = [0, 0]
    past_expedited_orders = [0, 0]
    q = dual_sourcing_controller.forward(current_inventory, past_regular_orders, past_expedited_orders)
    assert len(q) == 2
    assert q[0].shape == torch.Size([1, 1])
    assert q[1].shape == torch.Size([1, 1])

def test_dual_controller_get_total_cost(dual_sourcing_model: DualSourcingModel, dual_sourcing_controller: DualSourcingNeuralController):
    total_cost = dual_sourcing_controller.get_total_cost(dual_sourcing_model, sourcing_periods=5, seed=42)
    assert total_cost.item() == 12.0

def test_dual_controller_simulate(dual_sourcing_model: DualSourcingModel, dual_sourcing_controller: DualSourcingNeuralController):
    past_inventories, past_regular_orders, past_expedited_orders = dual_sourcing_controller.simulate(dual_sourcing_model, sourcing_periods=5, seed=42)
    assert past_inventories[0] == 10
    assert past_regular_orders[0] == 0
    assert past_expedited_orders[0] == 0

def test_dual_controller_train(dual_sourcing_model: DualSourcingModel, dual_sourcing_controller: DualSourcingNeuralController):
    dual_sourcing_controller.train(dual_sourcing_model, sourcing_periods=5, epochs=1)

def test_dual_controller_plot(dual_sourcing_model: DualSourcingModel, dual_sourcing_controller: DualSourcingNeuralController):
    dual_sourcing_controller.plot(dual_sourcing_model, sourcing_periods=5)